/**
 * This file contains auxiliary routines that are responsible for embeding and projecting an HSet to/from Poincare section.
 * Note, the integration od the infinite-dimensional ODE is performed in the full space, while the section fixes one of the coordinates.
 */ 

/// Remove first coordinate from a finite-dimensional vector 
template<class V>
V projectVector(const V& v){
  return V(v.dimension()-1,v.begin()+1);
}

// #####################################################################

/// Extend a finite-dimensional vector by adding zero as the first coordinate
template<class V>
V embedVector(const V& v){
  V r(v.dimension()+1);
  r[0] = 0.;
  for(unsigned i=0;i<v.dimension();++i)
    r[i+1] = v[i];
  return r;
}

// #####################################################################

/// Project finite-dimensional matrix by removing first row and first column
template<class M>
M projectMatrix(const M& m){
  int D = m.numberOfRows();
  M r(D-1,D-1);
  for(int i=1;i<D;++i)
    for(int j=1;j<D;++j)
      r(i,j) = m(i+1,j+1);
  return r;
}

// #####################################################################
/// Enlarge matrix by addind leading row and column that contain zeroes, except r(1,1)=1
template<class M>
M embedMatrix(const M& m){
  int D = m.numberOfRows();
  M r(D+1,D+1);
  r(1,1) = 1.;
  for(int i=1;i<=D;++i)
    for(int j=1;j<=D;++j)
      r(i+1,j+1) = m(i,j);
  return r;
}

// #####################################################################
/// Given a square matrix 'm', resize it to dimension 'dim' by adding 1 on the diagonal.
template<class M>
M resizeMatrix(const M& m, int dim){
  int i,j;
  int D = m.numberOfRows();
  M r(dim,dim);
  int d = capd::min(dim,D);
  for(i=1;i<=d;++i)
    for(j=1;j<=d;++j)
      r(i,j) = m(i,j);
  for(i=d+1;i<=dim;++i)
    r(i,i) = 1.;
  return r;
}

// #####################################################################
/// Given a vector u, resize it to dimension dim by adding zeroes to the tail
template<class V>
V resizeVector(const V& u, int dim){
  int d = capd::min(dim,(int)u.dimension());
  V r(dim);
  for(int i=0;i<d;++i)
    r[i] = u[i];
  return r;
}

// #####################################################################
/// Perform QR-decomopistion of Q and store the result again in Q.
template<class M>
void orthonormalizeColumns(M& Q){
  M A = Q;
  M R = Q;
  matrixAlgorithms::QR_decompose(A,Q,R);
}

// #####################################################################
/// Given a matrix Q, normalize its columns.
template<class M>
void normalizeColumns(M& Q){
  for(int i=0;i<Q.numberOfColumns();++i)
    Q.column(i).normalize();
}
